import { select } from "d3-selection";
import "d3-transition";
import { getTypeFromChartLayout, remToPx, getColorOrTexture, isTextRtl } from "./utils";
import {min, max} from "d3-array";
import { wrapStringToLines } from "@flourish/pocket-knife";
import { updatePatterns, updateCustomPatterns } from "./append-patterns-to";

function initUpdate(instance, state) {
	instance.update = function() {
		updatePatterns(state);
		const chart_layout = instance.chartLayout();
		const facet_name = instance.facetName();
		const offset = instance.offset();
		const line_height = 1.20;
		const data = [{ chart_layout, facet_name: facet_name}];
		const is_text_rtl = isTextRtl();

		const fg_label_class = "fg-label";
		const bg_label_class = "bg-label";

		const custom_patterns_set = new Set();

		var parseNum = instance.numberParser();
		var dur = instance.animationDuration();

		function getXTextSize() {
			return remToPx(state.x_font_size);
		}

		function getYTextSize() {
			return remToPx(state.y_font_size);
		}

		function splitLabelAndValue(string) {
			return string.replace(/\s*::\s*/g, "::").split("::").map(s => s.trim());
		}

		function getColors(highlight_color_str, state_highlight_color, state_text_color) {
			let colors = {
				highlight_color: state_highlight_color || "currentColor",
				text_color: state_text_color || "currentColor"
			};

			if (!highlight_color_str) return colors;

			if (highlight_color_str.includes("/")) {
				const split_colors = highlight_color_str.replace(/\s*\/\s*/g, "/").split("/").map(s => s.trim());
				colors.highlight_color = split_colors[0];
				colors.text_color = split_colors[1];
			}
			else {
				colors.highlight_color = highlight_color_str.trim();
				colors.text_color = state_text_color || "currentColor";
			}

			return colors;
		}

		function splitRange(string) {
			return string.replace(/\s*>>\s*/, ">>").split(">>").map(s => s.trim());
		}

		function getParsedValue(value, type, chart_layout, axis) {
			let parsed_value;
			if (type === "number") parsed_value = isNaN(parseNum(value)) ? parseNum(value) : value;
			else if (type === "datetime") {
				const parser = axis === "x" ? chart_layout.xDatetimeParse() : chart_layout.yDatetimeParse();
				parsed_value = parser(value);
			}
			else parsed_value = value;
			return parsed_value;
		}

		function valueIsWithinDomain(type, parsed_value, domain) {
			if (type === "number" || type === "datetime") {
				let sorted_domain = [...domain].sort((a, b) => a - b); // Pass in function so that dates are sorted correctly
				return ((parsed_value <= sorted_domain[1]) && (parsed_value >= sorted_domain[0]));
			}
			else {
				return domain.includes(parsed_value);
			}
		}

		function updateXLines(facet_group, data) {
			const chart_layout = data.chart_layout;
			let margin_left = chart_layout.margins.left;
			let margin_top = chart_layout.margins.top;
			let plot_height = Math.max(0, chart_layout.plot_height);
			let disabled = instance.xDisabled();

			let xScale = chart_layout.xScale();
			let type = getTypeFromChartLayout(chart_layout.xData());

			var line_data = [];

			if (!disabled && state.x_enabled && state.x_lines) {
				state.x_lines.split("\n").forEach(function(string) {
					var line = splitLabelAndValue(string);
					var label = line[0];
					var value = line[1];
					if (value === undefined) return;
					var colors = getColors(line[2], state.x_line_color, state.x_label_color);
					var facet = line[3] || null;
					if (facet && data.facet_name !== facet) return;

					const parsed_value = getParsedValue(value, type, chart_layout, "x");
					const domain = xScale.domain();
					const value_is_valid = valueIsWithinDomain(type, parsed_value, domain);

					if (value_is_valid) {
						line_data.push({
							label: label,
							x: parsed_value,
							line_color: colors.highlight_color,
							label_color: colors.text_color
						});
					}
				});
			}

			var lines = select(facet_group).selectAll("g.line.x").data(line_data);

			var lines_enter = lines.enter().append("g").attr("class", "line x")
				.attr("pointer-events", "none")
				.attr("transform", function(d) { return "translate(" + (xScale(d.x) || margin_left) + ", " + margin_top + ")"; });
			lines_enter.append("line");
			lines_enter.append("text");

			var lines_update = lines.merge(lines_enter);
			lines_update.transition().duration(dur)
				.attr("transform", function(d) { return "translate(" + (xScale(d.x) || margin_left) + ", " + margin_top + ")"; });
			lines_update.select("line")
				.attr("stroke-dasharray", state.x_line_dash + " " + state.x_line_dash)
				.attr("stroke", d => d.line_color)
				.attr("stroke-width", state.x_line_width)
				.attr("y2", plot_height);
			lines_update.select("text").each(function(d) {
				const x_coordinate = (xScale(d.x) || margin_left);
				const font_family = instance.fontFamily();
				const max_width_in_px = remToPx(state.x_label_max_width);
				const font_style = `${state.x_font_weight}  ${state.x_font_size} ${font_family}`;
				const wrapped_lines_array = wrapStringToLines(d.label,
					font_style,
					null,
					max_width_in_px);


				drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, bg_label_class);
				drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, fg_label_class);

				const num_of_lines = wrapped_lines_array.length;
				const {transform_str, text_anchor} = getTransformAndAnchorForX(this, xScale, plot_height, num_of_lines, x_coordinate);
				select(this)
					.attr("transform", transform_str)
					.attr("text-anchor", text_anchor);
			})
			.attr("font-size", getXTextSize())
			.attr("fill", d => d.label_color);

			lines_update.selectAll(`.${bg_label_class}`)
				.attr("display", state.x_shadow_enabled ? "block": "none")
				.attr("fill", "none")
				.attr("stroke", state.x_shadow_color)
				.attr("stroke-width", 2);

			lines.exit().remove();
		}

		function getXAreaWidth(range, plot_width, margin_left) {
			const min_val = min(range);
			const max_val = max(range);
			const min_val_or_y_axis = Math.max(min_val, margin_left);
			const end_of_chart = plot_width + margin_left;
			const max_val_or_end_of_chart = Math.min(max_val, end_of_chart);
			const width = Math.abs(max_val_or_end_of_chart - min_val_or_y_axis);
			const backup_width = 4;
			// if width is 0 (because user has entered one invalid value) then just show highlight of width 4 around the valid value
			return width || backup_width;
		}

		function getCustomPatternString(color, texture) {
			return JSON.stringify({
				color: color,
				texture: texture
			});
		}

		function getXAreas(data) {
			const chart_layout = data.chart_layout;
			let margin_left = chart_layout.margins.left;
			let margin_top = chart_layout.margins.top;
			let plot_width = Math.max(0, chart_layout.plot_width);
			let xScale = chart_layout.xScale();
			let type = getTypeFromChartLayout(data.chart_layout.xData());
			let disabled = instance.xDisabled();
			const texture = state.x_area_texture;
			var area_data = [];

			if (!disabled && state.x_enabled && state.x_areas) {
				state.x_areas.split("\n").map(function(string) {
					var area = splitLabelAndValue(string);
					if (area.length < 2) return;
					var label = area[0];
					var split_range = splitRange(area[1]);
					var range = split_range.map(function(value, i) {
						var parsed_value = getParsedValue(value, type, chart_layout, "x");
						var x_val = parsed_value ? xScale(parsed_value) : null;
						if (instance.xExtendByStep()) x_val += chart_layout.xMinStep()/2 * (i == 0 ? -1 : 1);
						return x_val;
					});
					var colors = getColors(area[2], state.x_area_color, state.x_label_color);
					const color = colors.highlight_color;
					let custom_pattern_i = null;
					if (area[2] !== undefined && color !== "currentColor" && texture !== "none") {
						const pattern_string = getCustomPatternString(color, texture);
						custom_patterns_set.add(pattern_string);
						custom_pattern_i = [...custom_patterns_set].indexOf(pattern_string);
					}
					var facet = area[3] || null;
					if (facet && data.facet_name !== facet) return;

					const lower_parsed_value = getParsedValue(split_range[0], type, chart_layout, "x");
					const upper_parsed_value = getParsedValue(split_range[1], type, chart_layout, "x");
					const domain = xScale.domain();
					const lower_value_is_valid = valueIsWithinDomain(type, lower_parsed_value, domain);
					const upper_value_is_valid = valueIsWithinDomain(type, upper_parsed_value, domain);

					if (lower_value_is_valid || upper_value_is_valid) {
						if (range.length > 1) area_data.push({
							label: label,
							range: range,
							area_color: color,
							label_color: colors.text_color,
							transform: `translate(${Math.max(min(range), margin_left)}, ${margin_top})`,
							width: getXAreaWidth(range, plot_width, margin_left),
							height: Math.max(0, chart_layout.plot_height),
							x: (min(range) || margin_left),
							custom_pattern_i: custom_pattern_i
						});
					}
				});
			}
			return area_data;
		}

		function updateXAreas(facet_group, area_data, scale) {
			var areas = select(facet_group).selectAll("g.area.x").data(area_data);
			var areas_enter = areas.enter().append("g")
				.attr("class", "area x")
				.attr("pointer-events", "none")
				.attr("transform", d => d.transform);

			areas_enter.append("rect");
			areas_enter.append("text");

			var areas_update = areas.merge(areas_enter);

			areas_update.transition().duration(dur)
				.attr("transform", d => d.transform);

			areas_update.select("rect")
				.attr("width", d => d.width)
				.style("fill", d => getColorOrTexture(d, state.x_area_texture, "x"))
				.attr("opacity", state.x_area_opacity)
				.attr("height", d => d.height);

			areas_update.select("text").each(function(d) {
				const font_family = instance.fontFamily();
				const max_width_in_px = remToPx(state.x_label_max_width);
				const font_style = `${state.x_font_weight} ${state.x_font_size} ${font_family}`;

				const wrapped_lines_array = wrapStringToLines(d.label,
					font_style,
					null,
					max_width_in_px);
				drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, bg_label_class);
				drawWrappedLabels(wrapped_lines_array, this, state.x_font_size, state.x_font_weight, fg_label_class);

				const num_of_lines = wrapped_lines_array.length;
				const {transform_str, text_anchor} = getTransformAndAnchorForX(this, scale, d.height, num_of_lines, d.x);
				select(this)
					.attr("transform", transform_str)
					.attr("text-anchor", text_anchor);
			})
			.attr("font-size", getXTextSize())
			.attr("fill", d => d.label_color);

			areas_update.selectAll(`.${bg_label_class}`)
				.attr("display", state.x_shadow_enabled ? "block": "none")
				.attr("fill", "none")
				.attr("stroke", state.x_shadow_color || "transparent")
				.attr("stroke-width", 2);

			areas.exit().remove();
		}

		function updateYLines(facet_group, data) {
			const chart_layout = data.chart_layout;
			let plot_width = Math.max(0, chart_layout.plot_width);
			let x = chart_layout.margins.left;
			let yScale = chart_layout.yScale();
			let type = getTypeFromChartLayout(chart_layout.yData());
			let disabled = instance.yDisabled();
			let margin_top = chart_layout.margins.top;

			var line_data = [];
			if (!disabled && state.y_enabled && state.y_lines) {
				state.y_lines.split("\n").forEach(function(string) {
					var line = splitLabelAndValue(string);
					var label = line[0];
					var value = line[1];
					if (value === undefined) return;
					var colors = getColors(line[2], state.y_line_color, state.y_label_color);
					var facet = line[3] || null;
					if (facet && data.facet_name !== facet) return;


					const parsed_value = getParsedValue(value, type, chart_layout, "u");
					const domain = yScale.domain();
					const value_is_valid = valueIsWithinDomain(type, parsed_value, domain);

					if (value_is_valid) {
						line_data.push({
							label: label,
							y: parsed_value,
							line_color: colors.highlight_color,
							label_color: colors.text_color
						});
					}
				});
			}

			var lines = select(facet_group).selectAll("g.line.y").data(line_data);

			var lines_enter = lines.enter().append("g").attr("class", "line y")
				.attr("pointer-events", "none")
				.attr("transform", function(d) { return "translate(" + x + ", " + (yScale(d.y) || 0) + ")"; });
			lines_enter.append("line").attr("width", 0);
			lines_enter.append("text");

			var lines_update = lines.merge(lines_enter);
			lines_update.transition().duration(dur)
				.attr("transform", function(d) { return "translate(" + x + ", " + (yScale(d.y) || 0) + ")"; });
			lines_update.select("line")
				.attr("stroke-dasharray", state.y_line_dash + " " + state.y_line_dash)
				.attr("stroke", d => d.line_color)
				.attr("stroke-width", state.y_line_width)
				.attr("x2", plot_width);
			lines_update.select("text")
				.each(function(d) {
					const font_family = instance.fontFamily();
					const y_coordinate = (yScale(d.y) || margin_top);
					const max_width_in_px = remToPx(state.y_label_max_width);
					const font_style = `${state.y_font_weight} ${state.y_font_size} ${font_family}`;
					const wrapped_lines_array = wrapStringToLines(d.label,
						font_style,
						null,
						max_width_in_px);

					drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, bg_label_class);
					drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, fg_label_class);

					const num_of_lines = wrapped_lines_array.length;
					const {transform_str, text_anchor} = getTransformAndAnchorForY(this, yScale, plot_width, num_of_lines, y_coordinate);

					select(this)
						.attr("transform", transform_str)
						.attr("text-anchor", text_anchor);
				})
				.attr("font-size", getYTextSize())
				.attr("fill", d => d.label_color);

			lines_update.selectAll(`.${bg_label_class}`)
				.attr("display", state.y_shadow_enabled ? "block": "none")
				.attr("fill", "none")
				.attr("stroke", state.y_shadow_color)
				.attr("stroke-width", 2);


			lines.exit().remove();
		}

		function getYAreaHeight(range, plot_height, margin_top) {
			const min_val = min(range);
			const max_val = max(range);
			const bottom_of_chart = plot_height + margin_top;
			const max_val_or_bottom_of_chart = Math.min(max_val, bottom_of_chart);
			const min_val_or_top_of_chart = Math.max(min_val, margin_top);
			const height = Math.abs(max_val_or_bottom_of_chart - min_val_or_top_of_chart);
			const backup_height = 4;
			// if height is 0 (because user has entered one invalid value) then just show highlight of height 4 around the valid value
			return height || backup_height;
		}

		function getYAreas(data) {
			const chart_layout = data.chart_layout;
			let margin_top = chart_layout.margins.top;
			let plot_height = Math.max(0, chart_layout.plot_height);
			let margin_left = chart_layout.margins.left;
			let yScale = chart_layout.yScale();
			let type = getTypeFromChartLayout(chart_layout.yData());
			let disabled = instance.yDisabled();
			const texture = state.y_area_texture;
			var area_data = [];

			if (!disabled && state.y_enabled && state.y_areas) {
				state.y_areas.split("\n").map(function(string) {
					var area = splitLabelAndValue(string);
					if (area.length < 2) return;
					var label = area[0];
					var split_range = splitRange(area[1]);
					var range = split_range.map(function(value, i) {
						var parsed_value = getParsedValue(value, type, chart_layout, "y");
						var y_val = parsed_value ? yScale(parsed_value) : parsed_value;
						if (instance.yExtendByStep()) y_val += chart_layout.yMinStep()/2 * (i == 0 ? 1 : -1);
						return y_val;
					});
					var colors = getColors(area[2], state.y_area_color, state.y_label_color);
					const color = colors.highlight_color;
					let custom_pattern_i = null;
					if (area[2] !== undefined && color !== "currentColor" && texture !== "none") {
						const pattern_string = getCustomPatternString(color, texture);
						custom_patterns_set.add(pattern_string);
						custom_pattern_i = [...custom_patterns_set].indexOf(pattern_string);
					}
					var facet = area[3] || null;
					if (facet && data.facet_name !== facet) return;

					const lower_parsed_value = getParsedValue(split_range[0], type, chart_layout, "y");
					const upper_parsed_value = getParsedValue(split_range[1], type, chart_layout, "y");
					const domain = yScale.domain();
					const lower_value_is_valid = valueIsWithinDomain(type, lower_parsed_value, domain);
					const upper_value_is_valid = valueIsWithinDomain(type, upper_parsed_value, domain);

					if (lower_value_is_valid || upper_value_is_valid) {
						if (range.length > 1) area_data.push({
							label: label,
							range: range,
							area_color: color,
							label_color: colors.text_color,
							width: Math.max(0, chart_layout.plot_width),
							height: getYAreaHeight(range, plot_height, margin_top),
							transform: `translate(${margin_left}, ${Math.max(min(range), margin_top)})`,
							y: (min(range) || margin_top),
							custom_pattern_i: custom_pattern_i
						});
					}
				});
			}
			return area_data;
		}

		function updateYAreas(facet_group, area_data, scale) {
			var areas = select(facet_group).selectAll("g.area.y").data(area_data);

			var areas_enter = areas.enter().append("g").attr("class", "area y")
				.attr("pointer-events", "none")
				.attr("transform", d => d.transform);

			areas_enter.append("rect");
			areas_enter.append("text");

			var areas_update = areas.merge(areas_enter);

			areas_update
				.transition().duration(dur)
				.attr("transform", d => d.transform);

			areas_update.select("rect")
				.attr("height", d => d.height)
				.attr("fill", d => getColorOrTexture(d, state.y_area_texture, "y"))
				.attr("opacity", state.y_area_opacity)
				.attr("width", d => d.width);

			areas_update.select("text").each(function(d) {
				const font_family = instance.fontFamily();
				const max_width_in_px = remToPx(state.y_label_max_width);
				const font_style = `${state.y_font_weight} ${state.y_font_size} ${font_family}`;

				const wrapped_lines_array = wrapStringToLines(d.label,
					font_style,
					null,
					max_width_in_px);

				drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, bg_label_class);
				drawWrappedLabels(wrapped_lines_array, this, state.y_font_size, state.y_font_weight, fg_label_class);
				const num_of_lines = wrapped_lines_array.length;
				const { transform_str, text_anchor } = getTransformAndAnchorForY(this, scale, d.width, num_of_lines, d.y);

				select(this)
					.attr("transform", transform_str)
					.attr("text-anchor", text_anchor);
			})
			.attr("font-size", getYTextSize())
			.attr("fill", d => d.label_color);

			areas_update.selectAll(`.${bg_label_class}`)
				.attr("display", state.y_shadow_enabled ? "block": "none")
				.attr("fill", "none")
				.attr("stroke", state.y_shadow_color || "transparent")
				.attr("stroke-width", 2);


			areas.exit().remove();
		}


		function drawWrappedLabels(wrapped_lines_array, target_el, font_size, font_weight, class_name) {
			var font_size_in_px = remToPx(font_size);
			var tspans = select(target_el).selectAll(`.${class_name}`).data(wrapped_lines_array);


			tspans.join("tspan")
					.text(function (d) { return d ; })
					.attr("class", class_name)
					.attr("y", function(_, i) {
						var y = (i * (font_size_in_px * line_height));
						return y + "px";
					})
					.attr("x", 0)
					.attr("font-weight", font_weight);

			tspans.exit().remove();
		}

		function getTransformAndAnchorForX(target_el, xScale, chart_height, num_of_lines, x_coordinate_of_highlight) {
			/* the reason we are working out the text anchor and the transform at
			the same time is because we only want to calculate if a label overlaps
			the chart once, and that affects both properties*/
			let transform_str;
			let x_position_of_text = 0;
			let text_anchor;

			const x_position_middle = state.x_label_x_position == "middle";
			const y_position_bottom = state.x_label_y_position == "bottom";
			const is_vertical = state.x_label_vertical;
			const line_height_in_px = getXTextSize() * line_height;
			// svg draws from the bottom of the first line of text, so we need
			// to times line height by num of lines minus one
			const text_block_offset = line_height_in_px * (num_of_lines - 1);

			const x_padding = getXTextSize() /2;
			const y_padding_if_horizontal = y_position_bottom ? text_block_offset + x_padding: line_height_in_px;
			const y_padding_if_vertical = x_padding;
			const y_padding = is_vertical ? y_padding_if_vertical : y_padding_if_horizontal;
			const vertical_center =  is_vertical ? (chart_height / 2) : (chart_height / 2) - (text_block_offset / 2);
			const max_x_value = xScale.range()[1];
			const min_x_value = xScale.range()[0];
			const text_width = target_el.getBoundingClientRect().width;

			const y_position_lookup = {
				top: y_padding,
				middle: vertical_center,
				bottom: chart_height - y_padding
			};
			const y_position = y_position_lookup[state.x_label_y_position];

			if (is_vertical) {
				const x_position_lookup = {
					right: getXTextSize(),
					middle: - (text_block_offset - x_padding)/2,
					left: - text_block_offset - x_padding
				};
				x_position_of_text = x_position_lookup[state.x_label_x_position];

				const text_anchor_lookup = {
					top: is_text_rtl ? "start" : "end",
					middle: "middle",
					bottom: is_text_rtl ? "end" : "start"
				};
				text_anchor = text_anchor_lookup[state.x_label_y_position];

				const overlap_at_end = x_coordinate_of_highlight + text_width + x_padding > max_x_value;

				const overlap_at_start = x_coordinate_of_highlight - text_width - x_padding < min_x_value;

				if (overlap_at_end) x_position_of_text = x_position_lookup.left;
				if (overlap_at_start) x_position_of_text = x_position_lookup.right;

				if (x_position_middle) {
					const overlap_at_end = x_coordinate_of_highlight + (text_width /2) > max_x_value;
					const overlap_at_start = x_coordinate_of_highlight - (text_width/2) < min_x_value;

					if (overlap_at_end) x_position_of_text = x_position_lookup.left;
					if (overlap_at_start) x_position_of_text = x_position_lookup.right;
				}

				transform_str = "translate(" + x_position_of_text + ", " + y_position + ") rotate(-90)";
			}
			else if (!is_vertical) {
				// Add padding depending on which side of the line the text is on
				// If there is not enough space for the label, it will move it to
				// the other side.

				const x_position_lookup = {
					right: x_padding,
					middle: 0,
					left: - x_padding
				};
				x_position_of_text = x_position_lookup[state.x_label_x_position];

				const text_anchor_lookup = {
					right: is_text_rtl ? "end": "start",
					middle: "middle",
					left: is_text_rtl ? "start" : "end"
				};
				text_anchor = text_anchor_lookup[state.x_label_x_position];

				let overlap_at_end;
				let overlap_at_start;

				if (x_position_middle) {
					overlap_at_end = x_coordinate_of_highlight + (text_width /2) > max_x_value;
					overlap_at_start = x_coordinate_of_highlight - (text_width/2) < min_x_value;
				}
				else {
					overlap_at_end = x_coordinate_of_highlight + text_width + x_padding > max_x_value;
					overlap_at_start = x_coordinate_of_highlight - text_width - x_padding < min_x_value;
				}

				if (overlap_at_end) {
					x_position_of_text = x_position_lookup.left;
					text_anchor = text_anchor_lookup.left;
				}
				if (overlap_at_start) {
					x_position_of_text = x_position_lookup.right;
					text_anchor = text_anchor_lookup.right;
				}

				transform_str = "translate(" + x_position_of_text + ", " + y_position + ")";
			}

			return {
				transform_str,
				text_anchor
			};
		}


		function getTransformAndAnchorForY(target_el, yScale, chart_width, num_of_lines, y_coordinate_of_line) {
			let text_anchor;
			let transform_str;

			const text_anchor_lookup = {
				left: is_text_rtl ? "end" : "start",
				middle: "middle",
				right: is_text_rtl ? "start": "end"
			};
			text_anchor = text_anchor_lookup[state.y_label_x_position];

			const line_height_in_px = getYTextSize() * line_height;
			const text_block_offset = line_height_in_px * (num_of_lines - 1);
			const x_padding = getYTextSize()/2;
			const text_offset_and_padding = text_block_offset + x_padding;
			const max_y_val = yScale.range()[0];
			const min_y_val = yScale.range()[1];
			const text_height = target_el.getBoundingClientRect().height;
			const y_position_middle = state.y_label_y_position == "middle";

			const y_position_lookup = {
				above: - text_offset_and_padding,
				middle: - (text_offset_and_padding/2) + x_padding,
				below: line_height_in_px
			};
			let y_position = y_position_lookup[state.y_label_y_position];

			const x_position_lookup = {
				right: chart_width - x_padding,
				middle: chart_width / 2,
				left: x_padding,
			};
			let x_position = x_position_lookup[state.y_label_x_position];

			let overlap_at_top;
			let overlap_at_bottom;

			if (y_position_middle) {
				overlap_at_top = y_coordinate_of_line - (text_height/2) < min_y_val;
				overlap_at_bottom = y_coordinate_of_line + (text_height/2) > max_y_val;
			}
			else {
				overlap_at_top = y_coordinate_of_line - text_height < min_y_val;
				overlap_at_bottom = y_coordinate_of_line + text_height > max_y_val;
			}

			if (overlap_at_top) y_position = y_position_lookup.below;
			if (overlap_at_bottom) y_position = y_position_lookup.above;

			transform_str = "translate(" + x_position + ", " + y_position + ")";

			return {
				transform_str,
				text_anchor
			};
		}


		function updateHighlights() {
			var parent_element = instance.appendTo();
			var axis_highlights = parent_element.selectAll("g.axis-highlights").data(data);
			var axis_highlights_enter = axis_highlights.enter().append("g").attr("class", "axis-highlights")
				.style("opacity", 0);
			axis_highlights.merge(axis_highlights_enter)
				.transition().duration(dur)
				.style("opacity", 1)
				.attr("transform", function() {
					if (!offset) return null;
					if (isNaN(offset.x) || isNaN(offset.y)) return null;
					return "translate(" + offset.x + ", " + offset.y + ")";
				})
				.each(function(d) {
					var facet_group = this;
					const x_areas = getXAreas(d);
					const y_areas = getYAreas(d);
					const custom_patterns = [...custom_patterns_set].map(c => JSON.parse(c));
					updateCustomPatterns(parent_element, custom_patterns);
					updateXAreas(facet_group, x_areas, d.chart_layout.xScale());
					updateYAreas(facet_group, y_areas, d.chart_layout.yScale());
					updateXLines(facet_group, d);
					updateYLines(facet_group, d);
				});

			axis_highlights.exit().remove();
		}

		updateHighlights();
	};
}


export { initUpdate };
